package net.wanji.opt.service.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.DateUnit;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.ObjectUtil;
import net.wanji.common.dto.CrossIdDTO;
import net.wanji.common.enums.CrossStatusEnum;
import net.wanji.common.enums.EventAlarmEnum;
import net.wanji.common.enums.EventAlarmSourceEnum;
import net.wanji.common.enums.TurnConvertEnum;
import net.wanji.common.utils.tool.CrossUtil;
import net.wanji.databus.dao.entity.*;
import net.wanji.databus.dao.mapper.*;
import net.wanji.databus.po.*;
import net.wanji.databus.vo.AbnormalCrossListVO;
import net.wanji.opt.bo.GreenwaveIdAndTimeStampBO;
import net.wanji.opt.bo.GreenwaveIdBO;
import net.wanji.opt.bo.SaveGreenwaveStrategyBO;
import net.wanji.opt.dao.mapper.CrossSchemeOptLogMapper;
import net.wanji.opt.dao.mapper.strategy.SceneStrategyIdeaMapper;
import net.wanji.opt.dao.mapper.strategy.StrategyMapper;
import net.wanji.opt.dao.mapper.trend.EventAlarmMapper;
import net.wanji.opt.dao.mapper.trend.GreenwaveInfoMapper;
import net.wanji.opt.dto.trend.AbnormalCrossListDTO;
import net.wanji.opt.dto.trend.GreenwaveListDTO;
import net.wanji.opt.po.base.CrossSchemeOptLogPO;
import net.wanji.opt.po.strategy.StrategyPO;
import net.wanji.opt.po.trend.EventAlarmPO;
import net.wanji.opt.service.TrendService;
import net.wanji.opt.vo.*;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @author Kent HAN
 * @date 2023/2/8 10:31
 */
@Service
public class TrendServiceImpl implements TrendService {
    private final GreenwaveInfoMapper greenwaveInfoMapper;
    private final BaseCrossInfoMapper baseCrossInfoMapper;
    private final CrossDataRealtimeMapper crossDataRealtimeMapper;
    private final CrossTurnDataRealtimeMapper crossTurnDataRealtimeMapper;
    private final CrossDirDataHistMapper crossDirDataHistMapper;
    private final EventAlarmMapper eventAlarmMapper;
    private final RidInfoMapper ridInfoMapper;
    private final BaseCrossDirInfoMapper baseCrossDirInfoMapper;
    private final CrossSchemeOptLogMapper crossSchemeOptLogMapper;
    private final GreenwaveCrossMapper greenwaveCrossMapper;
    private final CrossDirDataRealtimeMapper crossDirDataRealtimeMapper;
    private final GreenwaveHistMapper greenwaveHistMapper;
    private final GreenwaveRealtimeMapper greenwaveRealtimeMapper;
    private final StrategyMapper strategyMapper;
    private final BaseCrossSchedulesMapper baseCrossSchedulesMapper;
    private final BaseCrossSchedulesPlanMapper baseCrossSchedulesPlanMapper;
    private final BaseCrossSectionMapper baseCrossSectionMapper;
    private final BaseCrossSchemeMapper baseCrossSchemeMapper;
    private final BaseCrossPhaseMapper baseCrossPhaseMapper;
    private final BaseCrossPhaseLightsMapper crossPhaseLightsMapper;
    private final BaseCrossLaneLightsMapper baseCrossLaneLightsMapper;
    private final CrossBaseLaneInfoMapper crossBaseLaneInfoMapper;
    private final CrossTurnDataHistMapper crossTurnDataHistMapper;

    // 用于计算路口状态,key为方向，value为状态
    private Map<Integer, Integer> preStatus = new HashMap<Integer, Integer>() {{
        put(1, 1);
        put(2, 1);
        put(3, 1);
        put(4, 1);
        put(5, 1);
        put(6, 1);
        put(7, 1);
        put(8, 1);
    }};

    SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

    public TrendServiceImpl(GreenwaveInfoMapper greenwaveInfoMapper, BaseCrossInfoMapper baseCrossInfoMapper,
                            CrossDataRealtimeMapper crossDataRealtimeMapper,
                            CrossTurnDataRealtimeMapper crossTurnDataRealtimeMapper,
                            CrossDirDataHistMapper crossDirDataHistMapper, EventAlarmMapper eventAlarmMapper,
                            RidInfoMapper ridInfoMapper, BaseCrossDirInfoMapper baseCrossDirInfoMapper,
                            CrossSchemeOptLogMapper crossSchemeOptLogMapper, GreenwaveCrossMapper greenwaveCrossMapper, CrossDirDataRealtimeMapper crossDirDataRealtimeMapper, GreenwaveHistMapper greenwaveHistMapper, GreenwaveRealtimeMapper greenwaveRealtimeMapper, SceneStrategyIdeaMapper strategyIdeaMapper, StrategyMapper strategyMapper, BaseCrossSchedulesMapper baseCrossSchedulesMapper, BaseCrossSchedulesPlanMapper baseCrossSchedulesPlanMapper, BaseCrossSectionMapper baseCrossSectionMapper, BaseCrossSchemeMapper crossSchemeMapper, BaseCrossSchemeMapper baseCrossSchemeMapper, BaseCrossPhaseMapper baseCrossPhaseMapper, BaseCrossPhaseLightsMapper crossPhaseLightsMapper, BaseCrossLaneLightsMapper baseCrossLaneLightsMapper, CrossBaseLaneInfoMapper crossBaseLaneInfoMapper, CrossTurnDataHistMapper crossTurnDataHistMapper) {
        this.greenwaveInfoMapper = greenwaveInfoMapper;
        this.baseCrossInfoMapper = baseCrossInfoMapper;
        this.crossDataRealtimeMapper = crossDataRealtimeMapper;
        this.crossTurnDataRealtimeMapper = crossTurnDataRealtimeMapper;
        this.crossDirDataHistMapper = crossDirDataHistMapper;
        this.eventAlarmMapper = eventAlarmMapper;
        this.ridInfoMapper = ridInfoMapper;
        this.baseCrossDirInfoMapper = baseCrossDirInfoMapper;
        this.crossSchemeOptLogMapper = crossSchemeOptLogMapper;
        this.greenwaveCrossMapper = greenwaveCrossMapper;
        this.crossDirDataRealtimeMapper = crossDirDataRealtimeMapper;
        this.greenwaveHistMapper = greenwaveHistMapper;
        this.greenwaveRealtimeMapper = greenwaveRealtimeMapper;
        this.strategyMapper = strategyMapper;
        this.baseCrossSchedulesMapper = baseCrossSchedulesMapper;
        this.baseCrossSchedulesPlanMapper = baseCrossSchedulesPlanMapper;
        this.baseCrossSectionMapper = baseCrossSectionMapper;
        this.baseCrossSchemeMapper = baseCrossSchemeMapper;
        this.baseCrossPhaseMapper = baseCrossPhaseMapper;
        this.crossPhaseLightsMapper = crossPhaseLightsMapper;
        this.baseCrossLaneLightsMapper = baseCrossLaneLightsMapper;
        this.crossBaseLaneInfoMapper = crossBaseLaneInfoMapper;
        this.crossTurnDataHistMapper = crossTurnDataHistMapper;
    }

    @Override
    public List<GreenwaveListVO> greenwaveList(GreenwaveListDTO greenwaveListDTO) {
        Integer status = greenwaveListDTO.getStatus();
        String name = greenwaveListDTO.getName();
        Integer type = greenwaveListDTO.getType();
        List<GreenwaveListVO> greenwaveListVOList = greenwaveInfoMapper.listGreenwave(status, name, type);
        // 3、4都算拥堵
        if (status != null && status == 3) {
            List<GreenwaveListVO> extraList = greenwaveInfoMapper.listGreenwave(4, name, type);
            greenwaveListVOList.addAll(extraList);
        }
        changeStatus(greenwaveListVOList);
        for (GreenwaveListVO greenwaveListVO : greenwaveListVOList) {
            // 计算持续时间
            buildDuration(greenwaveListVO);
            // 构造路口列表
            Integer greenwaveId = greenwaveListVO.getId();
            List<GreenwaveListVO.CrossListElement> crossList = buildCrossList(greenwaveId);
            greenwaveListVO.setCrossList(crossList);
        }
        return greenwaveListVOList;
    }

    @Override
    public GreenwaveStats buildGreenwaveStats(List<GreenwaveListVO> greenwaveListVOList) {
        GreenwaveStats greenwaveStats = new GreenwaveStats();
        Integer smooth = greenwaveStats.getSmooth();
        Integer slow = greenwaveStats.getSlow();
        Integer congestion = greenwaveStats.getCongestion();
        for (GreenwaveListVO greenwaveListVO : greenwaveListVOList) {
            Integer realtimeStatus = greenwaveListVO.getRealtimeStatus();
            if (realtimeStatus == 1) {
                smooth += 1;
            } else if (realtimeStatus == 2) {
                slow += 1;
            } else if (realtimeStatus == 3) {
                congestion += 1;
            }
        }
        greenwaveStats.setSmooth(smooth);
        greenwaveStats.setSlow(slow);
        greenwaveStats.setCongestion(congestion);
        return greenwaveStats;
    }

    @Override
    public AbnormalCrossVO abnormalCrossList(AbnormalCrossListDTO abnormalCrossListDTO) {
        AbnormalCrossVO abnormalCrossVO = new AbnormalCrossVO();
        // 构建问题路口列表
        Integer status = abnormalCrossListDTO.getStatus();
        String name = abnormalCrossListDTO.getName();
        Integer type = abnormalCrossListDTO.getType();
        List<AbnormalCrossListVO> abnormalCrossListVOList =
                crossDataRealtimeMapper.selectAbnormalCross(status, name, type);
        // 添加优化状态
        for (AbnormalCrossListVO abnormalCrossListVO : abnormalCrossListVOList) {
            abnormalCrossListVO.setOptStatus(2); // 2未优化
            String crossId = abnormalCrossListVO.getId();
            // 查询优化日志
            List<CrossSchemeOptLogPO> crossSchemeOptLogPOS = crossSchemeOptLogMapper.selectMaxByCrossId(crossId);
            if (CollectionUtil.isNotEmpty(crossSchemeOptLogPOS)) {
                CrossSchemeOptLogPO crossSchemeOptLogPO = crossSchemeOptLogPOS.get(0);
                Date endTime = crossSchemeOptLogPO.getEndTime();
                if (ObjectUtil.isEmpty(endTime)) {
                    abnormalCrossListVO.setOptStatus(1); // 1优化中
                }
            }
        }
        // 坐标格式转换
        for (AbnormalCrossListVO abnormalCrossListVO : abnormalCrossListVOList) {
            String locationStr = abnormalCrossListVO.getLocationStr();
            double[] lonLat = CrossUtil.getLonLat(locationStr);
            List<Double> location = new ArrayList<>();
            location.add(lonLat[0]);
            location.add(lonLat[1]);
            abnormalCrossListVO.setLocation(location);
            // 关联路口信息
            Integer realtimeStatus = abnormalCrossListVO.getRealtimeStatus();
            String endCrossId = abnormalCrossListVO.getId();
            if (Objects.equals(realtimeStatus, CrossStatusEnum.SPILLOVER.getCode())) {
                // 溢出只返回溢出方向的关联路口
                String spilloverDirs = abnormalCrossListVO.getSpilloverDirs();
                String[] split = spilloverDirs.split(",");
                List<Integer> spilloverDirList = new ArrayList<>();
                for (String s : split) {
                    spilloverDirList.add(Integer.parseInt(s));
                }
                buildRidData(abnormalCrossListVO, endCrossId, spilloverDirList);
            } else if (Objects.equals(realtimeStatus, CrossStatusEnum.CONGESTION.getCode())){
                // 拥堵返回所有方向关联路口
                List<Integer> inDirs = baseCrossDirInfoMapper.selectInDirsByCrossId(endCrossId);
                buildRidData(abnormalCrossListVO, endCrossId, inDirs);
            }
        }
        CrossComparator crossComparator = new CrossComparator();
        abnormalCrossListVOList.sort(crossComparator);
        abnormalCrossVO.setAbnormalCrossList(abnormalCrossListVOList);
        // 构造统计信息
        List<AbnormalCrossListVO> listForStats =
                crossDataRealtimeMapper.selectAbnormalCross(null, null, null);
        buildAbnormalCrossStats(abnormalCrossVO, listForStats);

        return abnormalCrossVO;
    }

    private static class CrossComparator implements Comparator<AbnormalCrossListVO> {
        @Override
        public int compare(AbnormalCrossListVO o1, AbnormalCrossListVO o2) {
            // 先按路口状态排序
            Integer o1status = o1.getRealtimeStatus();
            Integer o2status = o2.getRealtimeStatus();
            int flag = o2status - o1status;
            if (flag == 0) {
                // 再按拥堵时长排序
                Long o1duration = o1.getDuration();
                Long o2Duration = o2.getDuration();
                long res = o2Duration - o1duration;
                return (int) res;
            }
            return flag;
        }
    }

    private void buildRidData(AbnormalCrossListVO abnormalCrossListVO, String endCrossId,
                              List<Integer> spilloverDirList) {
        List<List<Double>> ridLocationList = new ArrayList<>();
        List<String> ridWktList = new ArrayList<>();
        for (Integer spilloverDir : spilloverDirList) {
            RidInfoEntity ridInfo = ridInfoMapper.selectByEndInDir(endCrossId, spilloverDir);
            if (ridInfo != null) {
                // 关联路口坐标
                String startCrossId = ridInfo.getStartCrossId();
                TBaseCrossInfo crossInfoPO = baseCrossInfoMapper.selectByPrimaryKey(startCrossId);
                if (crossInfoPO != null) {
                    String startLocationStr = crossInfoPO.getLocation();
                    double[] startLonLat = CrossUtil.getLonLat(startLocationStr);
                    List<Double> startLocation = new ArrayList<>();
                    startLocation.add(startLonLat[0]);
                    startLocation.add(startLonLat[1]);
                    ridLocationList.add(startLocation);
                }
                // 关联路口wkt
                String wkt = ridInfo.getWkt();
                ridWktList.add(wkt);
            }
        }
        abnormalCrossListVO.setRidLocationList(ridLocationList);
        abnormalCrossListVO.setRidWktList(ridWktList);
    }

    @Override
    public AbnormalCrossDetailVO abnormalCrossDetail(CrossIdDTO crossIdDTO) {
        String crossId = crossIdDTO.getCrossId();
        AbnormalCrossDetailVO abnormalCrossDetailVO = new AbnormalCrossDetailVO();

        // 构造路口实时数据
        CrossDataRealtimePO crossDataRealtimePO = crossDataRealtimeMapper.selectByCrossId(crossId);
        abnormalCrossDetailVO.setCrossStatus(crossDataRealtimePO.getStatus());
        abnormalCrossDetailVO.setCrossQueueLength(crossDataRealtimePO.getQueueLength());
        abnormalCrossDetailVO.setCrossFlowRate((double)crossDataRealtimePO.getFlow());
        abnormalCrossDetailVO.setCrossStopTimes(crossDataRealtimePO.getStopTimes());
        abnormalCrossDetailVO.setCongestionIndex(crossDataRealtimePO.getCongestionIndex());
        // 构造转向实时数据
        List<AbnormalCrossDetailVO.TurnDataElement> turnData = buildTurnData(crossId);
        abnormalCrossDetailVO.setTurnData(turnData);
        // 构造方向历史数据
        List<AbnormalCrossDetailVO.DirDataElement> dirData = buildDirData(crossId);
        abnormalCrossDetailVO.setDirData(dirData);

        return abnormalCrossDetailVO;
    }

    @Override
    public List<EventAlarmVO> eventAlarm(Date todayTime) throws ParseException {
        List<EventAlarmPO> eventAlarmPOList = eventAlarmMapper.selectByTodayTime(todayTime);
        List<EventAlarmVO> eventAlarmOutVOList = new ArrayList<>();
        for (EventAlarmPO eventAlarmPO : eventAlarmPOList) {
            EventAlarmVO eventAlarmOutVO = new EventAlarmVO();
            eventAlarmOutVO.setEventDesc(eventAlarmPO.getEventDesc());
            // 事件类型
            Integer typeCode = eventAlarmPO.getType();
            String msgByCode = EventAlarmEnum.getMsgByCode(typeCode);
            eventAlarmOutVO.setType(msgByCode);

            Date endTime = eventAlarmPO.getEndTime();
            if (typeCode == 1 || typeCode == 2) {
                // 发生时间
                Date startTime = eventAlarmPO.getStartTime();
                String startTimeStr = sdf.format(startTime);
                String substring = startTimeStr.substring(11, 16);
                eventAlarmOutVO.setStartTime(substring);
                // 时长
                if (endTime == null) {
                    Date nowTime = DateUtil.date();
                    long between = DateUtil.between(startTime, nowTime, DateUnit.MINUTE);
                    int betweenInt = (int) between;
                    eventAlarmOutVO.setDuration(betweenInt);
                } else {
                    long between = DateUtil.between(startTime, endTime, DateUnit.MINUTE);
                    int betweenInt = (int) between;
                    eventAlarmOutVO.setDuration(betweenInt);
                }
            }
            if (typeCode == 3 || typeCode == 4) {
                // 时间范围
                Date startTime = eventAlarmPO.getStartTime();
                String startTimeStr = sdf.format(startTime);
                String startTimeSubStr = startTimeStr.substring(0, 16); // 截取到分钟
                String endTimeStr = sdf.format(endTime);
                String endTimeSubStr = endTimeStr.substring(0, 16); // 截取到分钟
                eventAlarmOutVO.setTimeSpan(startTimeSubStr + "~" + endTimeSubStr);
            }

            Integer sourceCode = eventAlarmPO.getSource();
            String sourceStr = EventAlarmSourceEnum.getMsgByCode(sourceCode);
            eventAlarmOutVO.setSource(sourceStr);
            eventAlarmOutVO.setCoordinate(eventAlarmPO.getWkt());

            eventAlarmOutVOList.add(eventAlarmOutVO);

        }
        return eventAlarmOutVOList;
    }

    @Override
    public List<GreenWaveCrossMonitorVO> greenwaveCrossMonitor(GreenwaveIdBO greenwaveIdBO) {
        Integer greenwaveId = greenwaveIdBO.getGreenwaveId();
        List<GreenwaveCrossPO> greenwaveCrossPOList = greenwaveCrossMapper.selectByGreenwaveId(greenwaveId);
        List<GreenWaveCrossMonitorVO> res = buildRes(greenwaveCrossPOList);
        return res;
    }

    @Override
    public List<GreenwaveRunMonitorVO> greenwaveRunMonitor(GreenwaveIdBO greenwaveIdBO) {
        List<GreenwaveRunMonitorVO> res = new ArrayList<>();
        Date nowTime = new Date();
        // todo 测试用，固定当前时间
        nowTime = DateUtil.parse("2023-06-09 15:00:00","yyyy-MM-dd HH:mm:ss");
        List<GreenwaveHistPOExt> extList = greenwaveHistMapper.selectRunMonitor(nowTime);
        for (GreenwaveHistPOExt greenwaveHistPOExt : extList) {
            GreenwaveRunMonitorVO greenWaveRunMonitorVO = new GreenwaveRunMonitorVO();
            String hour = greenwaveHistPOExt.getHour();
            String hourSub = hour.substring(11, 16); // 截取小时
            greenWaveRunMonitorVO.setHour(hourSub);
            greenWaveRunMonitorVO.setAvgSpeed(greenwaveHistPOExt.getAvgSpeed());
            greenWaveRunMonitorVO.setCongestionIndex(greenwaveHistPOExt.getCongestionIndex());
            res.add(greenWaveRunMonitorVO);
        }
        return res;
    }

    @Override
    public GreenwaveDetailVO greenwaveDetail(GreenwaveIdBO greenwaveIdBO) {
        Integer greenwaveId = greenwaveIdBO.getGreenwaveId();
        GreenwaveDetailVO res = new GreenwaveDetailVO();
        GreenwaveRealtimePO greenwaveRealtimePO = greenwaveRealtimeMapper.selectById(greenwaveId);
        GreenwaveInfoPO greenwaveInfoPO = greenwaveInfoMapper.selectById(greenwaveId);
        res.setName(greenwaveInfoPO.getName());
        res.setStatus(greenwaveRealtimePO.getStatus());

        // 绿波带宽
        double width = greenwaveInfoPO.getWidth();
        double designSpeed = greenwaveInfoPO.getDesignSpeed() / 3.6; // 换算成米/秒
        int widthSeconds = (int)(width / designSpeed);
        Integer dir = greenwaveInfoPO.getDir();
        if (dir == 0) { // 正向
            res.setForwardWidth(widthSeconds);
        } else if (dir == 1) { // 反向
            res.setReverseWidth(widthSeconds);
        } else if (dir == 2) { // 双向
            res.setForwardWidth(widthSeconds);
            res.setReverseWidth(widthSeconds);
        }

        double uncoordinatePhaseQueue = greenwaveRealtimePO.getUncoordinatePhaseQueue();
        res.setUncoordinatePhaseQueue((int)uncoordinatePhaseQueue);
        double noparkPassRate = greenwaveRealtimePO.getNoparkPassRate();
        res.setNoparkPassRate((int) (noparkPassRate * 100));
        res.setTrvalTime(greenwaveRealtimePO.getTrvalTime());
        // 获取策略信息
        List<String> strategyList = strategyMapper.selectAllNames();
        res.setStrategyList(strategyList);
        List<StrategyPO> strategyPOList = strategyMapper.selectByGreenwaveId(greenwaveId);
        List<String> collect = strategyPOList.stream()
                .map(StrategyPO::getStrategyName)
                .collect(Collectors.toList());
        res.setStrategySelected(collect);
        // 绿波路口信息
        List<GreenwaveDetailVO.GreenwaveCross> greenwaveCrossList = buildGreenwaveCrossList(greenwaveId);
        res.setGreenwaveCrossList(greenwaveCrossList);
        // 绿波带
        if (dir == 0) { // 正向
            res.setForwardGreenwaveList(buildForwardGreenwaveList(greenwaveCrossList, greenwaveInfoPO));
        } else if (dir == 1) { // 反向
            res.setReverseGreenwaveList(buildReverseGreenwaveList(greenwaveCrossList, greenwaveInfoPO));
        } else if (dir == 2) { // 双向
            res.setForwardGreenwaveList(buildForwardGreenwaveList(greenwaveCrossList, greenwaveInfoPO));
            res.setReverseGreenwaveList(buildReverseGreenwaveList(greenwaveCrossList, greenwaveInfoPO));
        }

        return res;
    }

    private List<GreenwaveDetailVO.GreenwavePoint> buildReverseGreenwaveList(
            List<GreenwaveDetailVO.GreenwaveCross> greenwaveCrossList, GreenwaveInfoPO greenwaveInfoPO) {
        List<GreenwaveDetailVO.GreenwavePoint> res = new ArrayList<>();
        // 找关键路口
        GreenwaveDetailVO.GreenwaveCross keyCross = null;
        Integer keyCrossBeginSeconds = null;
        for (GreenwaveDetailVO.GreenwaveCross greenwaveCross : greenwaveCrossList) {
            if (1 == greenwaveCross.getIsKeyRoute()) {
                keyCross = greenwaveCross;
                GreenwaveDetailVO.GreenwavePoint greenwavePoint = buildGreenwavePoint(greenwaveCross);
                Integer coordSeconds = greenwavePoint.getCoordSeconds();
                greenwavePoint.setCoordOffset(coordSeconds);
                res.add(greenwavePoint);
                keyCrossBeginSeconds = calcCoordBeginSeconds(greenwaveCross) + coordSeconds;
            }
        }
        Integer sort = keyCross.getSort();
        // 从关键路口向上
        int upSize = sort - 2;
        Integer preCrossBeginSeconds = keyCrossBeginSeconds;
        for (int i = upSize; i >= 0; i--) {
            GreenwaveDetailVO.GreenwaveCross greenwaveCross = greenwaveCrossList.get(i);
            GreenwaveDetailVO.GreenwavePoint greenwavePoint = buildGreenwavePoint(greenwaveCross);
            // 计算行程时间
            int distanceToNextCross = greenwaveCross.getDistanceToNextCross();
            double designSpeed = greenwaveInfoPO.getDesignSpeed() / 3.6; // 换算成米/秒
            int travalTime = (int)(distanceToNextCross / designSpeed);
            // 计算协调相位绿灯开始的绝对时间
            int absoluteSecondsBegin = calcCoordBeginSeconds(greenwaveCross);
            // 计算绿波点位的绝对时间
            int absoluteSecondsPoint = preCrossBeginSeconds + travalTime;
            greenwavePoint.setCoordOffset(absoluteSecondsPoint - absoluteSecondsBegin);

            res.add(greenwavePoint);
            preCrossBeginSeconds = absoluteSecondsPoint;
        }
        // 从关键路口向下
        preCrossBeginSeconds = keyCrossBeginSeconds;
        for (int i = sort; i < greenwaveCrossList.size(); i++) {
            GreenwaveDetailVO.GreenwaveCross greenwaveCross = greenwaveCrossList.get(i);
            GreenwaveDetailVO.GreenwaveCross greenwaveCrossAbove = greenwaveCrossList.get(i - 1);
            GreenwaveDetailVO.GreenwavePoint greenwavePoint = buildGreenwavePoint(greenwaveCross);
            // 计算行程时间
            int distanceToNextCross = greenwaveCrossAbove.getDistanceToNextCross();
            double designSpeed = greenwaveInfoPO.getDesignSpeed() / 3.6; // 换算成米/秒
            int travalTime = (int)(distanceToNextCross / designSpeed);
            // 计算协调相位绿灯开始的绝对时间
            int absoluteSecondsBegin = calcCoordBeginSeconds(greenwaveCross);
            // 计算绿波点位的绝对时间
            int absoluteSecondsPoint = preCrossBeginSeconds - travalTime;
            greenwavePoint.setCoordOffset(absoluteSecondsPoint - absoluteSecondsBegin);

            res.add(greenwavePoint);
            preCrossBeginSeconds = absoluteSecondsPoint;
        }
        List<GreenwaveDetailVO.GreenwavePoint> sorted = res.stream()
                .sorted(Comparator.comparing(GreenwaveDetailVO.GreenwavePoint::getSort))
                .collect(Collectors.toList());
        return sorted;
    }

    private List<GreenwaveDetailVO.GreenwavePoint> buildForwardGreenwaveList(
            List<GreenwaveDetailVO.GreenwaveCross> greenwaveCrossList, GreenwaveInfoPO greenwaveInfoPO) {
        List<GreenwaveDetailVO.GreenwavePoint> res = new ArrayList<>();
        // 找关键路口
        GreenwaveDetailVO.GreenwaveCross keyCross = null;
        Integer keyCrossBeginSeconds = null;
        for (GreenwaveDetailVO.GreenwaveCross greenwaveCross : greenwaveCrossList) {
            if (1 == greenwaveCross.getIsKeyRoute()) {
                keyCross = greenwaveCross;
                GreenwaveDetailVO.GreenwavePoint greenwavePoint = buildGreenwavePoint(greenwaveCross);
                greenwavePoint.setCoordOffset(0);
                res.add(greenwavePoint);
                keyCrossBeginSeconds = calcCoordBeginSeconds(greenwaveCross);
            }
        }
        Integer sort = keyCross.getSort();
        // 从关键路口向上
        int upSize = sort - 2;
        Integer preCrossBeginSeconds = keyCrossBeginSeconds;
        for (int i = upSize; i >= 0; i--) {
            GreenwaveDetailVO.GreenwaveCross greenwaveCross = greenwaveCrossList.get(i);
            GreenwaveDetailVO.GreenwavePoint greenwavePoint = buildGreenwavePoint(greenwaveCross);
            // 计算行程时间
            int distanceToNextCross = greenwaveCross.getDistanceToNextCross();
            double designSpeed = greenwaveInfoPO.getDesignSpeed() / 3.6; // 换算成米/秒
            int travalTime = (int)(distanceToNextCross / designSpeed);
            // 计算协调相位绿灯开始的绝对时间
            int absoluteSecondsBegin = calcCoordBeginSeconds(greenwaveCross);
            // 计算绿波点位的绝对时间
            int absoluteSecondsPoint = preCrossBeginSeconds - travalTime;
            greenwavePoint.setCoordOffset(absoluteSecondsPoint - absoluteSecondsBegin);

            res.add(greenwavePoint);
            preCrossBeginSeconds = absoluteSecondsPoint;
        }
        // 从关键路口向下
        preCrossBeginSeconds = keyCrossBeginSeconds;
        for (int i = sort; i < greenwaveCrossList.size(); i++) {
            GreenwaveDetailVO.GreenwaveCross greenwaveCross = greenwaveCrossList.get(i);
            GreenwaveDetailVO.GreenwaveCross greenwaveCrossAbove = greenwaveCrossList.get(i - 1);
            GreenwaveDetailVO.GreenwavePoint greenwavePoint = buildGreenwavePoint(greenwaveCross);
            // 计算行程时间
            int distanceToNextCross = greenwaveCrossAbove.getDistanceToNextCross();
            double designSpeed = greenwaveInfoPO.getDesignSpeed() / 3.6; // 换算成米/秒
            int travalTime = (int)(distanceToNextCross / designSpeed);
            // 计算协调相位绿灯开始的绝对时间
            int absoluteSecondsBegin = calcCoordBeginSeconds(greenwaveCross);
            // 计算绿波点位的绝对时间
            int absoluteSecondsPoint = preCrossBeginSeconds + travalTime;
            greenwavePoint.setCoordOffset(absoluteSecondsPoint - absoluteSecondsBegin);

            res.add(greenwavePoint);
            preCrossBeginSeconds = absoluteSecondsPoint;
        }
        List<GreenwaveDetailVO.GreenwavePoint> sorted = res.stream()
                .sorted(Comparator.comparing(GreenwaveDetailVO.GreenwavePoint::getSort))
                .collect(Collectors.toList());
        return sorted;
    }

    private int calcCoordBeginSeconds(GreenwaveDetailVO.GreenwaveCross greenwaveCross) {
        int res = 0;
        List<GreenwaveDetailVO.GreenwavePhase> greenwavePhaseList = greenwaveCross.getGreenwavePhaseList();
        for (GreenwaveDetailVO.GreenwavePhase greenwavePhase : greenwavePhaseList) {
            int seconds = greenwavePhase.getSeconds();
            int isCoordinate = greenwavePhase.getIsCoordinate();
            if (1 != isCoordinate) {
                res += seconds;
            } else {
                break;
            }
        }
        return res;
    }

    private GreenwaveDetailVO.GreenwavePoint buildGreenwavePoint(GreenwaveDetailVO.GreenwaveCross greenwaveCross) {
        GreenwaveDetailVO.GreenwavePoint greenwavePoint = new GreenwaveDetailVO.GreenwavePoint();
        greenwavePoint.setSort(greenwaveCross.getSort());
        greenwavePoint.setCrossId(greenwaveCross.getCrossId());
        greenwavePoint.setCrossName(greenwaveCross.getCrossName());
        List<GreenwaveDetailVO.GreenwavePhase> greenwavePhaseList = greenwaveCross.getGreenwavePhaseList();
        for (GreenwaveDetailVO.GreenwavePhase greenwavePhase : greenwavePhaseList) {
            if (1 == greenwavePhase.getIsCoordinate()) {
                greenwavePoint.setCoordSeconds(greenwavePhase.getSeconds());
            }
        }
        return greenwavePoint;
    }

    private List<GreenwaveDetailVO.GreenwaveCross> buildGreenwaveCrossList(Integer greenwaveId) {
        List<GreenwaveDetailVO.GreenwaveCross> res = new ArrayList<>();
        List<GreenwaveCrossPO> greenwaveCrossList = greenwaveCrossMapper.selectByGreenwaveId(greenwaveId);
        // 相对相位差列表
        List<Integer> relativeOffsets = greenwaveCrossList.stream()
                .map(GreenwaveCrossPO::getOffset)
                .collect(Collectors.toList());
        // 绝对相位差列表
        List<Integer> absoluteOffsets = buildAbsoluteOffsets(relativeOffsets);

        int size = greenwaveCrossList.size();
        for (int i = 0; i < size; i++) {
            GreenwaveCrossPO greenwaveCrossPO = greenwaveCrossList.get(i);
            GreenwaveDetailVO.GreenwaveCross greenwaveCross = new GreenwaveDetailVO.GreenwaveCross();
            String crossId = greenwaveCrossPO.getCrossId();
            greenwaveCross.setCrossId(crossId);
            BaseCrossInfoPO baseCrossInfoPO = baseCrossInfoMapper.selectById(crossId);
            if (baseCrossInfoPO == null) {
                throw new RuntimeException("无此路口基础信息");
            }
            greenwaveCross.setCrossName(baseCrossInfoPO.getName());
            greenwaveCross.setIsKeyRoute(greenwaveCrossPO.getIsKeyRoute());
            greenwaveCross.setSort(greenwaveCrossPO.getSort());
            // 到下一个路口距离
            Integer distanceToNextCross = null;
            if (i < size - 1) { // 不是最后一个路口
                double nextCrossLen = greenwaveCrossPO.getNextCrossLen();
                distanceToNextCross = (int)nextCrossLen;
            }
            greenwaveCross.setDistanceToNextCross(distanceToNextCross);
            // 计算平均速度
            Double speed = calcSpeed(greenwaveCrossPO, crossId);
            greenwaveCross.setSpeed(speed);
            // 获取当前方案
            Integer sectionId = greenwaveCrossPO.getSectionId();
            CrossSectionPO crossSectionPO = baseCrossSectionMapper.selectById(sectionId);
            if (crossSectionPO == null) {
                throw new RuntimeException("无当前时段方案信息");
            }
            Integer currentSchemeId = crossSectionPO.getSchemeId();
            BaseCrossSchemePO baseCrossSchemePO = baseCrossSchemeMapper.selectById(currentSchemeId);
            greenwaveCross.setOffset(baseCrossSchemePO.getOffset());
            // 构造相位时差图
            Integer absoluteOffset = absoluteOffsets.get(i);
            Integer crossSort = greenwaveCrossPO.getSort();
            List<GreenwaveDetailVO.GreenwavePhase> greenwavePhaseList =
                    getGreenwavePhaseList(baseCrossSchemePO, absoluteOffset, crossSort);
            greenwaveCross.setGreenwavePhaseList(greenwavePhaseList);

            res.add(greenwaveCross);
        }
        return res;
    }

    private List<Integer> buildAbsoluteOffsets(List<Integer> relativeOffsets) {
        List<Integer> res = new ArrayList<>();
        for (int i = 0; i < relativeOffsets.size(); i++) {
            Integer subSum = 0;
            for (int j = 0; j <= i; j++) {
                subSum += relativeOffsets.get(j);
            }
            res.add(subSum);
        }
        return res;
    }

    private List<GreenwaveDetailVO.GreenwavePhase> getGreenwavePhaseList(BaseCrossSchemePO baseCrossSchemePO,
                                                                         Integer absoluteOffset, Integer crossSort) {
        List<GreenwaveDetailVO.GreenwavePhase> res = new ArrayList<>();
        Integer schemeId = baseCrossSchemePO.getId();
        String crossId = baseCrossSchemePO.getCrossId();
        Integer coordPhaseId = baseCrossSchemePO.getCoordPhase();
        // 获取方案相位并按相位序排序
        List<CrossPhasePO> crossPhasePOList = baseCrossPhaseMapper.selectByCrossIdAndSchemeId(crossId, schemeId);
        if (1 == crossSort) { // 第一个路口没有相位差
            int size = crossPhasePOList.size();
            for (int i = 0; i < size; i++) {
                CrossPhasePO crossPhasePO = crossPhasePOList.get(i);
                GreenwaveDetailVO.GreenwavePhase greenwavePhase = new GreenwaveDetailVO.GreenwavePhase();

                greenwavePhase.setSort(i + 1);
                greenwavePhase.setPhaseNo(crossPhasePO.getPhaseNo());
                greenwavePhase.setSeconds(crossPhasePO.getPhaseTime());
                Integer phaseId = crossPhasePO.getId();
                greenwavePhase.setIsCoordinate(ObjectUtil.equals(phaseId, coordPhaseId) ? 1 : 0);

                res.add(greenwavePhase);
            }
            return res;
        } else {
            // 其余路口，首先根据绝对相位差构造第一个相位，与最后一个相位首位相接
            GreenwaveDetailVO.GreenwavePhase firstGreenwavePhase = new GreenwaveDetailVO.GreenwavePhase();
            int size = crossPhasePOList.size();
            CrossPhasePO endCrossPhasePO = crossPhasePOList.get(size - 1);
            firstGreenwavePhase.setSort(1);
            firstGreenwavePhase.setPhaseNo(endCrossPhasePO.getPhaseNo());
            firstGreenwavePhase.setSeconds(absoluteOffset);
            Integer firstPhaseId = endCrossPhasePO.getId();
            firstGreenwavePhase.setIsCoordinate(ObjectUtil.equals(firstPhaseId, coordPhaseId) ? 1 : 0);
            res.add(firstGreenwavePhase);

            for (int i = 0; i < size; i++) {
                CrossPhasePO crossPhasePO = crossPhasePOList.get(i);
                GreenwaveDetailVO.GreenwavePhase greenwavePhase = new GreenwaveDetailVO.GreenwavePhase();

                greenwavePhase.setSort(i + 2);
                greenwavePhase.setPhaseNo(crossPhasePO.getPhaseNo());
                if (i == size - 1){
                    greenwavePhase.setSeconds(crossPhasePO.getPhaseTime() - absoluteOffset);
                } else {
                    greenwavePhase.setSeconds(crossPhasePO.getPhaseTime());
                }
                Integer phaseId = crossPhasePO.getId();
                greenwavePhase.setIsCoordinate(ObjectUtil.equals(phaseId, coordPhaseId) ? 1 : 0);

                res.add(greenwavePhase);
            }
            return res;
        }
    }

    private int convertWeekDay(int i) {
        if (1 == i) {
            return 7;
        }
        return i - 1;
    }

    private Double calcSpeed(GreenwaveCrossPO greenwaveCrossPO, String currentCrossId) {
        Integer outDir = greenwaveCrossPO.getOutDir();
        CrossDirDataRealtimePO crossDirDataRealtimePO =
                crossDirDataRealtimeMapper.selectByCrossIdAndDirType(currentCrossId, outDir);
        return crossDirDataRealtimePO.getSpeed();
    }

    @Override
    public List<GreenwaveCrossMetricsVO> greenwaveCrossMetrics(GreenwaveIdAndTimeStampBO greenwaveIdAndTimeStampBO) {
        List<GreenwaveCrossMetricsVO> res = new ArrayList<>();

        Integer greenwaveId = greenwaveIdAndTimeStampBO.getGreenwaveId();
        Long timeStamp = greenwaveIdAndTimeStampBO.getTimeStamp();
        // todo 测试用，时间戳设定为 1676082600
        timeStamp = 1676082600L;

        List<GreenwaveCrossPO> greenwaveCrossList = greenwaveCrossMapper.selectByGreenwaveId(greenwaveId);
        int crossListSize = greenwaveCrossList.size();
        int lastCrossIndex = crossListSize - 1;
        for (int i = 0; i < crossListSize; i++) {
            GreenwaveCrossPO greenwaveCrossPO = greenwaveCrossList.get(i);
            GreenwaveCrossMetricsVO greenwaveCrossMetricsVO = new GreenwaveCrossMetricsVO();
            String crossId = greenwaveCrossPO.getCrossId();
            greenwaveCrossMetricsVO.setCrossId(crossId);
            BaseCrossInfoPO baseCrossInfoPO = baseCrossInfoMapper.selectById(crossId);
            if (baseCrossInfoPO != null) {
                greenwaveCrossMetricsVO.setCrossName(baseCrossInfoPO.getName());
            }
            String currentCrossId = crossId;
            long currentSeconds = timeStamp;
            long preSeconds = currentSeconds - 60 * 5;
            // 获取当前方案
            Integer sectionId = greenwaveCrossPO.getSectionId();
            CrossSectionPO crossSectionPO = baseCrossSectionMapper.selectById(sectionId);
            Integer schemeId = crossSectionPO.getSchemeId();
            // 获取协调相位对应的车道集合
            BaseCrossSchemePO baseCrossSchemePO = baseCrossSchemeMapper.selectById(schemeId);
            Integer coordPhaseID = baseCrossSchemePO.getCoordPhase();
            List<CrossPhaseLightsPO> crossPhaseLightsPOList = crossPhaseLightsMapper.selectByPhaseId(coordPhaseID);
            List<Integer> lightIds = crossPhaseLightsPOList.stream()
                    .map(CrossPhaseLightsPO::getLightsId)
                    .collect(Collectors.toList());
            List<CrossLaneLightsPO> crossLaneLightsPOList = baseCrossLaneLightsMapper.selectByLightIds(lightIds);
            List<String> laneIds = crossLaneLightsPOList.stream()
                    .map(CrossLaneLightsPO::getLaneId)
                    .collect(Collectors.toList());
            List<CrossBaseLaneInfoPO> lanePOList = crossBaseLaneInfoMapper.selectBatchIds(laneIds);
            // 构造方向+转向列表
            List<List<Integer>> dirTurnList = new ArrayList<>();
            for (CrossBaseLaneInfoPO lanePO : lanePOList) {
                Integer dir = lanePO.getDir();
                Integer turn = lanePO.getTurn();
                List<Integer> dirTurn = new ArrayList<>();
                dirTurn.add(dir);
                dirTurn.add(turn);
                dirTurnList.add(dirTurn);
            }
            // 计算指标
            List<CrossTurnDataHistPO> listForNoPark = new ArrayList<>();
            for (List<Integer> dirTurn : dirTurnList) {
                Integer dir = dirTurn.get(0);
                Integer turnInt = dirTurn.get(1);
                String turn = TurnConvertEnum.getCodeByKey(turnInt);
                CrossTurnDataHistPO noPark =
                        crossTurnDataHistMapper.selectNoPark(currentCrossId, dir, turn, currentSeconds, preSeconds);
                if (noPark != null) {
                    listForNoPark.add(noPark);
                }
            }
            List<CrossTurnDataHistPO> allPO =
                    crossTurnDataHistMapper.selectByCrossId(currentCrossId, currentSeconds, preSeconds);
            List<CrossTurnDataHistPO> listForPhaseQueue = getPhaseQueueList(listForNoPark, allPO);
            setMetrics(greenwaveCrossMetricsVO, greenwaveCrossPO, listForNoPark, listForPhaseQueue);

            res.add(greenwaveCrossMetricsVO);
        }
        return res;
    }

    private List<CrossTurnDataHistPO> getPhaseQueueList(
            List<CrossTurnDataHistPO> listForNoPark, List<CrossTurnDataHistPO> allPO) {
        Iterator<CrossTurnDataHistPO> iterator = allPO.iterator();
        while (iterator.hasNext()) {
            CrossTurnDataHistPO next = iterator.next();
            for (CrossTurnDataHistPO crossTurnDataHistPO : listForNoPark) {
                if (Objects.equals(next.getTurnType(), crossTurnDataHistPO.getTurnType())
                        && Objects.equals(next.getInDir(), crossTurnDataHistPO.getInDir())) {
                    iterator.remove();
                }
            }
        }
        return allPO;
    }

    @Override
    @Transactional
    public void saveGreenwaveStrategy(SaveGreenwaveStrategyBO saveGreenwaveStrategyBO) {
        Integer greenwaveId = saveGreenwaveStrategyBO.getGreenwaveId();
        List<String> strategyNames = saveGreenwaveStrategyBO.getStrategyNames();
        strategyMapper.deleteGreenwave(greenwaveId);
        for (String strategyName : strategyNames) {
            // 获取策略ID
            StrategyPO strategyPO = strategyMapper.selectByName(strategyName);
            Integer strategyId = strategyPO.getId();
            // 更新绿波策略关系
            strategyMapper.insertGreenwave(greenwaveId, strategyId);
        }
    }

    private void setMetrics(GreenwaveCrossMetricsVO greenwaveCrossMetricsVO, GreenwaveCrossPO greenwaveCrossPO,
                            List<CrossTurnDataHistPO> listForNoPark, List<CrossTurnDataHistPO> listForPhaseQueue) {
        if (CollectionUtil.isNotEmpty(listForNoPark) && CollectionUtil.isNotEmpty(listForPhaseQueue)) {
            Integer noparkPassRate = getAveRate(listForNoPark);
            greenwaveCrossMetricsVO.setNoparkPassRate(noparkPassRate);
            Integer uncoordinatePhaseQueue = getPhaseQueue(listForPhaseQueue);
            greenwaveCrossMetricsVO.setUncoordinatePhaseQueue(uncoordinatePhaseQueue);
            Double length = greenwaveCrossPO.getNextCrossLen();
            Double speed = getSpeed(listForNoPark);
            double travelTime = length / speed;
            greenwaveCrossMetricsVO.setTrvalTime((int)travelTime);
        }
    }

    private Double getSpeed(List<CrossTurnDataHistPO> listForNoPark) {
        Double total = 0.0;
        for (CrossTurnDataHistPO crossDirDataHistPO : listForNoPark) {
            Double speed = crossDirDataHistPO.getSpeed();
            total += speed;
        }
        double avgSpeed = total / listForNoPark.size();
        return avgSpeed;
    }

    private Integer getPhaseQueue(List<CrossTurnDataHistPO> listForPhaseQueue) {
        double maxQueue = 0.0;
        for (CrossTurnDataHistPO crossDirDataHistPO : listForPhaseQueue) {
            double queueLength = crossDirDataHistPO.getQueueLength();
            if (queueLength > maxQueue) {
                maxQueue = queueLength;
            }
        }
        return (int)maxQueue;
    }

    private Integer getAveRate(List<CrossTurnDataHistPO> crossDirDataHistPOList) {
        Double total = 0.0;
        for (CrossTurnDataHistPO crossDirDataHistPO : crossDirDataHistPOList) {
            Double noStopRate = crossDirDataHistPO.getNoStopRate();
            total += noStopRate;
        }
        int size = crossDirDataHistPOList.size();
        double avg = total / size;
        return (int)(avg * 100);
    }

    private List<GreenWaveCrossMonitorVO> buildRes(List<GreenwaveCrossPO> greenwaveCrossPOList) {
        List<GreenWaveCrossMonitorVO> res = new ArrayList<>();
        for (GreenwaveCrossPO greenwaveCrossPO : greenwaveCrossPOList) {
            GreenWaveCrossMonitorVO greenWaveCrossMonitorVO = new GreenWaveCrossMonitorVO();

            String crossId = greenwaveCrossPO.getCrossId();
            CrossDataRealtimePO crossDataRealtimePO = crossDataRealtimeMapper.selectByCrossId(crossId);
            BaseCrossInfoPO baseCrossInfoPO = baseCrossInfoMapper.selectById(crossId);
            if (crossDataRealtimePO != null && baseCrossInfoPO != null) {
                greenWaveCrossMonitorVO.setCrossId(crossId);
                greenWaveCrossMonitorVO.setCrossName(baseCrossInfoPO.getName());
                greenWaveCrossMonitorVO.setFlow(crossDataRealtimePO.getFlow());

                List<CrossDirDataRealtimePO> crossDirDataRealtimePOList =
                        crossDirDataRealtimeMapper.selectBycrossId(crossId);
                Integer capacity = 0;
                for (CrossDirDataRealtimePO crossDirDataRealtimePO : crossDirDataRealtimePOList) {
                    capacity += crossDirDataRealtimePO.getCapacity();
                }
                greenWaveCrossMonitorVO.setCapacity(capacity);

                greenWaveCrossMonitorVO.setCongestionIndex(crossDataRealtimePO.getCongestionIndex());
                res.add(greenWaveCrossMonitorVO);
            }
        }
        return res;
    }

    private List<AbnormalCrossDetailVO.DirDataElement> buildDirData(String crossId) {
        initPreStatus();
        List<AbnormalCrossDetailVO.DirDataElement> dirData = new ArrayList<>();
        // 获取当前时间之前40分钟的10位时间戳
        long currentSeconds = DateUtil.currentSeconds();
        // todo 测试，当前时间固定为 1676082536
        currentSeconds = 1676082536;
        long preSeconds = currentSeconds - 40 * 60;
        List<CrossDirDataHistPO> crossDirDataHistPOList =
                crossDirDataHistMapper.selectByCrossIdAndTimestamp(crossId, preSeconds);
        List<Integer> dirs = baseCrossDirInfoMapper.selectInDirsByCrossId(crossId);

        Date currentMinute = DateUtil.date(currentSeconds * 1000);
        Date startMinute = DateUtil.offsetMinute(currentMinute, -59);

        for (int i = 1; i <= 60; i++) {
            // 下一分钟
            Date nextMinute = DateUtil.offsetMinute(startMinute, 1);
            for (Integer dir : dirs) {
                AbnormalCrossDetailVO.DirDataElement dirDataElement = new AbnormalCrossDetailVO.DirDataElement();
                dirDataElement.setDir(dir);
                dirDataElement.setTime(startMinute);
                int status = calcStatus(crossDirDataHistPOList, startMinute, nextMinute, dir);
                dirDataElement.setStatus(status);
                dirData.add(dirDataElement);
            }
            startMinute = nextMinute;
        }

        return dirData;
    }

    private void initPreStatus() {
        for (Integer key : preStatus.keySet()) {
            preStatus.put(key, 1);
        }
    }

    /**
     * 计算路口状态
     *
     * @author Kent HAN
     * @date 2023/6/20 9:07
     */
    private Integer calcStatus(List<CrossDirDataHistPO> crossDirDataHistPOList, Date preMinute, Date currentMinute,
                               int dir) {
        for (CrossDirDataHistPO crossDirDataHistPO : crossDirDataHistPOList) {
            int dirType = crossDirDataHistPO.getDirType();
            long batchTime = crossDirDataHistPO.getBatchTime();
            long preSecond = preMinute.getTime() / 1000;
            long currentSecond = currentMinute.getTime() / 1000;
            if ( dirType == dir && batchTime <= currentSecond && batchTime >= preSecond) {
                Integer status = crossDirDataHistPO.getStatus();
                // 转换状态，3、4都算拥堵
                if (status == 4) {
                    status = 3;
                }
                preStatus.put(dir, status);
            }
        }
        return preStatus.get(dir);
    }

    private List<AbnormalCrossDetailVO.TurnDataElement> buildTurnData(String crossId) {
        List<AbnormalCrossDetailVO.TurnDataElement> turnData = new ArrayList<>();
        List<CrossTurnDataRealtimePO> crossTurnDataRealtimePOList = crossTurnDataRealtimeMapper.selectByCrossId(crossId);
        Map<Integer, List<CrossTurnDataRealtimePO>> collect = crossTurnDataRealtimePOList.stream()
                .collect(Collectors.groupingBy(CrossTurnDataRealtimePO::getInDir));
        for (Map.Entry<Integer, List<CrossTurnDataRealtimePO>> entry : collect.entrySet()) {
            AbnormalCrossDetailVO.TurnDataElement turnDataElement = new AbnormalCrossDetailVO.TurnDataElement();
            Integer dir = entry.getKey();
            turnDataElement.setDir(dir);
            // 构造转向列表
            List<CrossTurnDataRealtimePO> subList = entry.getValue();
            List<AbnormalCrossDetailVO.TurnDataElement.TurnListElement> turnList = buildTurnList(subList);
            turnDataElement.setTurnList(turnList);
            
            turnData.add(turnDataElement);
        }
        return turnData;
    }

    private List<AbnormalCrossDetailVO.TurnDataElement.TurnListElement> buildTurnList(
            List<CrossTurnDataRealtimePO> subList) {
        List<AbnormalCrossDetailVO.TurnDataElement.TurnListElement> turnList = new ArrayList<>();

        for (CrossTurnDataRealtimePO crossTurnDataRealtimePO : subList) {
            AbnormalCrossDetailVO.TurnDataElement.TurnListElement turnListElement =
                    new AbnormalCrossDetailVO.TurnDataElement.TurnListElement();

            turnListElement.setTurn(crossTurnDataRealtimePO.getTurnType());
            turnListElement.setTurnQueueLength(crossTurnDataRealtimePO.getQueueLength());
            turnListElement.setTurnStopTimes(crossTurnDataRealtimePO.getStopTimes());
            turnListElement.setFlow(crossTurnDataRealtimePO.getFlow());

            turnList.add(turnListElement);
        }
        return turnList;
    }

    private static void buildAbnormalCrossStats(AbnormalCrossVO abnormalCrossVO,
                                                List<AbnormalCrossListVO> abnormalCrossListVOList) {
        AbnormalCrossStats abnormalCrossStats = new AbnormalCrossStats();
        Integer normal = abnormalCrossStats.getNormal();
        Integer unbalance = abnormalCrossStats.getUnbalance();
        Integer congestion = abnormalCrossStats.getCongestion();
        Integer spillover = abnormalCrossStats.getSpillover();
        for (AbnormalCrossListVO abnormalCrossListVO : abnormalCrossListVOList) {
            Integer realtimeStatus = abnormalCrossListVO.getRealtimeStatus();
            if (realtimeStatus == 0) {
                normal += 1;
            } else if (realtimeStatus == 1) {
                unbalance += 1;
            } else if (realtimeStatus == 2) {
                congestion += 1;
            } else if (realtimeStatus == 3) {
                spillover += 1;
            }
        }
        abnormalCrossStats.setNormal(normal);
        abnormalCrossStats.setUnbalance(unbalance);
        abnormalCrossStats.setCongestion(congestion);
        abnormalCrossStats.setSpillover(spillover);
        abnormalCrossVO.setAbnormalCrossStats(abnormalCrossStats);
    }

    private static void changeStatus(List<GreenwaveListVO> greenwaveListVOList) {
        if (greenwaveListVOList != null) {
            for (GreenwaveListVO greenwaveListVO : greenwaveListVOList) {
                if (greenwaveListVO.getRealtimeStatus() == 4) {
                    greenwaveListVO.setRealtimeStatus(3);
                }
            }
        }
    }

    private List<GreenwaveListVO.CrossListElement> buildCrossList(Integer greenwaveId) {
        List<GreenwaveListVO.CrossListElement> crossList = new ArrayList<>();
        List<String> crossIdList = greenwaveInfoMapper.selectCrossIdsById(greenwaveId);
        if (CollectionUtil.isNotEmpty(crossIdList)) {
            List<BaseCrossInfoPO> baseCrossInfoPOList = baseCrossInfoMapper.selectByCrossIds(crossIdList);
            for (BaseCrossInfoPO baseCrossInfoPO : baseCrossInfoPOList) {
                GreenwaveListVO.CrossListElement crossListElement = new GreenwaveListVO.CrossListElement();
                crossListElement.setCrossId(baseCrossInfoPO.getId());
                crossListElement.setName(baseCrossInfoPO.getName());
                List<Double> location = buildLocation(baseCrossInfoPO);
                crossListElement.setLocation(location);

                crossList.add(crossListElement);
            }
        }
        return crossList;
    }

    private List<Double> buildLocation(BaseCrossInfoPO baseCrossInfoPO) {
        String location = baseCrossInfoPO.getLocation();
        double[] lonLat = CrossUtil.getLonLat(location);
        List<Double> res = new ArrayList<>();
        res.add(lonLat[0]);
        res.add(lonLat[1]);
        return res;
    }

    private static void buildDuration(GreenwaveListVO greenwaveListVO) {
        Date startTime = greenwaveListVO.getStartTime();
        Date endTime = greenwaveListVO.getEndTime();
        Long duration = null;
        if (endTime != null) {
            duration = DateUtil.between(startTime, endTime, DateUnit.MINUTE);
        } else {
            duration = DateUtil.between(startTime, new Date(), DateUnit.MINUTE);
        }
        greenwaveListVO.setDuration(duration);
    }
}
